#ifndef TRANSFORMMATRIX_H
#define TRANSFORMMATRIX_H

#include <valarray>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <cmath>
#include <cassert>

#ifndef size_t
typedef unsigned int size_t;
#endif

using namespace std;

/**
 * @brief This class provides the functionality of a 3D geometry/projection 
 * transform matrix assuming a right-handed coordinate system (RHS).
 * <b> Since it uses assertions for checking index boundaries define the
 * NDEBUG flag (-DNDEBUG) for release versions! </b> 
 * @author Ruediger Knoerig
 */
 
template<class T> class TransformMatrix{
public:
	/** default constructor, results in a unity matrix */
	TransformMatrix() : matrixdata(16) { for(size_t k=0; k<16;k+=5) matrixdata[k]=T(1.0); }

	/**
	 * copy constructor
	 */
    TransformMatrix(const TransformMatrix<T> &src) : matrixdata(src.matrixdata) {}
    
	/** destructor */
	virtual ~TransformMatrix() { }
	
	/** 
	 * method resetting the matrix to the unity matrix 
	 */
	void reset() { matrixdata=0; for(size_t k=0; k<16;k+=5) matrixdata[k]=T(1.0);  } 	

	/**
	 * @brief inversion method
	 * Sets the matrix to its inverse. Algorithm after intel's AP 962.
	 */
	void invert();
		
	/**
	 * @brief method for transforming a 3-element vector.
	 * The w component is assumed to be 1 and will be (implicitly) 1 afterwards.
	 * @param v vector to be transformed. The vector class must have at 
	 * least 3 elements which must be accessible via the []-operator.
	 * @param vt reference to the 3-element target vector. Is \f$v'=v\cdot\mathbf{A}\f$ 
	 * afterwards.
	 */
	template<class Vector> inline void transform(const Vector &v,Vector &vt) const;
	
	/**
	 * @brief method for transforming a 3-element vector the fast way.
	 * It assumes that the transform matrix doesn't change the w component of
	 * the vector <b> so don't use this method in conjunction with the perspective
	 * view transform! </b>
	 * @param v vector to be transformed. The vector class must have at 
	 * least 3 elements which must be accessible via the []-operator.
	 * @param vt reference to the target vector. Is \f$v'=v\cdot\mathbf{A}\f$ 
	 * afterwards.
	 */
	template<class Vector> inline void transformQuick(const Vector &v,Vector &vt) const;
		
	/**
	 * @brief method for transforming a 4-element vector (homogenous coordinates). 
	 * @param v vector to be transformed. The vector class must have at 
	 * least 4 elements which must be accessible via the []-operator.
	 * @param vt reference to the target vector. Is \f$v'=v\cdot\mathbf{A}\f$ 
	 * afterwards.
	 */
	template<class Vector> inline void transformHomogen(const Vector &v,Vector &vt) const;
	
	/**
	 * @brief method applying a translation to the matrix.
	 * @param x x offset
	 * @param y y offset
	 * @param z z offset
	 */
	inline void translate(T x, T y, T z);
	
	/**
	 * @brief method applying a rotation around the z axis
	 * @param phi angle in radians
	 */
	inline void rotateAroundZ(T phi);
	
	/**
	 * @brief method applying a rotation around the x axis
	 * @param phi angle in radians
	 */
	inline void rotateAroundX(T phi);
	
	/**
	 * @brief method applying a rotation around the y axis
	 * @param phi angle in radians
	 */
	inline void rotateAroundY(T phi);
	
	/**
	 * @brief method applying a scale operation
	 * @param sx scaling in x direction
	 * @param sy scaling in y direction
	 * @param sz scaling in z direction
	 */
	inline void scale(T sx, T sy, T sz);
	
	/**
	 * @brief apply an orthogonal projection transformation (view from bottom)
	 */
	inline void projectionOrthoBottom();
	
	/**
	 * @brief apply an orthogonal projection transformation (view from top)
	 */
	inline void projectionOrthoTop();
	
	/**
	 * @brief apply an orthogonal projection transformation (view from left)
	 */
	inline void projectionOrthoLeft();	
	
	/**
	 * @brief apply an orthogonal projection transformation (view from right)
	 */
	inline void projectionOrthoRight();
		
	/**
	 * @brief applies an one-point perspectivic view transformation.
	 * @param d focal length of the virtual camera / distance of the virtual 
	 * observer from the view plane
	 */
	inline void projectionOnePointPersp(T d);
	
	/**
	 * @brief access operator for an element value
	 * @param row row number of the element
	 * @param column column number of the element 
	 * @return copy of the element value
	 */
	T operator()(size_t row,size_t column) const { assert(row < 4); assert(column <4 );  return matrixdata[column+4*row]; }
	
	/**
	 * access operator for an element value
	 * @param row row number of the element
	 * @param column column number of the element 
	 * @return copy of the element value
	 */
	T &operator()(size_t row,size_t column) { assert(row < 4); assert(column <4 );  return matrixdata[column+4*row]; }
	
	/**
	 * comparison for equality
	 * @param other matrix to compare with
	 * @return true if all elements in the matrices are the same
	 */
	bool operator==(const TransformMatrix<T> &other) const { for(size_t k=0; k<16;k++) if(matrixdata[k]!=other.matrixdata[k]) return false; return true; }
	
	/**
	 * comparison for inequality
	 * @param other matrix to compare with
	 * @return true if not all elements in the matrices are the same
	 */
	bool operator!=(const TransformMatrix<T> &other) { for(size_t k=0; k<16;k++) if(matrixdata[k]!=other.matrixdata[k]) return true; return false;  }
	
	/**
	 * add-to operator (element-wise addition)
	 * @param other matrix which should be added 
	 */
	void operator+=(const TransformMatrix<T> &other) { matrixdata += other.matrixdata; }
	
	/**
	 * subtract-from operator (element-wise substraction)
	 * @param other matrix which should substracted
	 */
	void operator-=(const TransformMatrix<T> &other) { matrixdata -= other.matrixdata; }
	
	/**
	 * multiply-with operator (matrix multiplication).
	 * @param b multiplicant
	 */
	inline void operator*=(const TransformMatrix<T> &b); 
	
	/**
	 * @brief matrix scaling operation (element-wise multiplication with a scalar)
	 * \f$\left\{a_{ij}\right\}=\left\{a_{ij}\cdot b\right\}\;\forall i,j \f$
	 * @param scale scalar value to multiply the elements with)
	 */
	void operator*=(T scale) { matrixdata *= scale; }
	 	
private: // private attributes
	/** internal representation of the matrix */
	valarray<T> matrixdata;
};


/**
 * @brief addition operator
 * @param a first summand
 * @param b second summand
 * @return sum a+b
 */
template<class T> TransformMatrix<T> operator+(const TransformMatrix<T> &a,const TransformMatrix<T> &b)
{
	TransformMatrix<T> c(a);
	c+=b;
	return c;
}
 
/**
 * @brief subtraction operator
 * @param a minuend
 * @param b subtrahend 
 * @return difference a-b
 */
template<class T> TransformMatrix<T> operator-(const TransformMatrix<T> &a,const TransformMatrix<T> &b)
{
	TransformMatrix<T> c(a);
	c-=b;
	return c;
}

/**
 * @brief matrix multiplication operator
 * @param a first multiplicator
 * @param b second multiplicator
 * @return product a*b
 */
template<class T> TransformMatrix<T> operator*(const TransformMatrix<T> &a,const TransformMatrix<T> &b)
{
	TransformMatrix<T> c(a);
	c*=b;
	return c;
}

/**
 * scaling operator (multiplication with a scalar)
 * @param a matrix
 * @param b scalar multiplicator
 * @return product \f$\left\{a_{ij}\right\}=\left\{a_{ij}\cdot b\right\}\;\forall i,j \f$
 */
template<class T> TransformMatrix<T> operator*(const TransformMatrix<T> &a,T b)
{
	TransformMatrix<T> c(a);
	c*=b;
	return b;
}

/**
 * multiplication operator (multiplication with a vector)
 * @param v Vector to multiply with
 * @param b scalar multiplicator
 * @return product \f$\left\{a_{ij}\right\}=\left\{a_{ij}\cdot b\right\}\;\forall i,j \f$
 */
/*template<class T> template<class Vector> TransformMatrix<T> operator*(const Vector &v,const TransformMatrix<T> &A)
{
	Vector vt(v);
	A.transform(v,vt);
	return vt;
}*/

template<class T> void TransformMatrix<T>::operator*=(const TransformMatrix<T> &b)
{
	TransformMatrix<T> a(*this);
	matrixdata[0]=a.matrixdata[0]*b.matrixdata[0]+a.matrixdata[1]*b.matrixdata[4]+a.matrixdata[2]*b.matrixdata[8]+a.matrixdata[3]*b.matrixdata[12];
	matrixdata[1]=a.matrixdata[0]*b.matrixdata[1]+a.matrixdata[1]*b.matrixdata[5]+a.matrixdata[2]*b.matrixdata[9]+a.matrixdata[3]*b.matrixdata[13];
	matrixdata[2]=a.matrixdata[0]*b.matrixdata[2]+a.matrixdata[1]*b.matrixdata[6]+a.matrixdata[2]*b.matrixdata[10]+a.matrixdata[3]*b.matrixdata[14];
	matrixdata[3]=a.matrixdata[0]*b.matrixdata[3]+a.matrixdata[1]*b.matrixdata[7]+a.matrixdata[2]*b.matrixdata[11]+a.matrixdata[3]*b.matrixdata[15];
	
	matrixdata[4]=a.matrixdata[4]*b.matrixdata[0]+a.matrixdata[5]*b.matrixdata[4]+a.matrixdata[6]*b.matrixdata[8]+a.matrixdata[7]*b.matrixdata[12];
	matrixdata[5]=a.matrixdata[4]*b.matrixdata[1]+a.matrixdata[5]*b.matrixdata[5]+a.matrixdata[6]*b.matrixdata[9]+a.matrixdata[7]*b.matrixdata[13];
	matrixdata[6]=a.matrixdata[4]*b.matrixdata[2]+a.matrixdata[5]*b.matrixdata[6]+a.matrixdata[6]*b.matrixdata[10]+a.matrixdata[7]*b.matrixdata[14];
	matrixdata[7]=a.matrixdata[4]*b.matrixdata[3]+a.matrixdata[5]*b.matrixdata[7]+a.matrixdata[6]*b.matrixdata[11]+a.matrixdata[7]*b.matrixdata[15];
	
	matrixdata[8]=a.matrixdata[8]*b.matrixdata[0]+a.matrixdata[9]*b.matrixdata[4]+a.matrixdata[10]*b.matrixdata[8]+a.matrixdata[11]*b.matrixdata[12];
	matrixdata[9]=a.matrixdata[8]*b.matrixdata[1]+a.matrixdata[9]*b.matrixdata[5]+a.matrixdata[10]*b.matrixdata[9]+a.matrixdata[11]*b.matrixdata[13];
	matrixdata[10]=a.matrixdata[8]*b.matrixdata[2]+a.matrixdata[9]*b.matrixdata[6]+a.matrixdata[10]*b.matrixdata[10]+a.matrixdata[11]*b.matrixdata[14];
	matrixdata[11]=a.matrixdata[8]*b.matrixdata[3]+a.matrixdata[9]*b.matrixdata[7]+a.matrixdata[10]*b.matrixdata[11]+a.matrixdata[11]*b.matrixdata[15];
	
	matrixdata[12]=a.matrixdata[12]*b.matrixdata[0]+a.matrixdata[13]*b.matrixdata[4]+a.matrixdata[14]*b.matrixdata[8]+a.matrixdata[15]*b.matrixdata[12];
	matrixdata[13]=a.matrixdata[12]*b.matrixdata[1]+a.matrixdata[13]*b.matrixdata[5]+a.matrixdata[14]*b.matrixdata[9]+a.matrixdata[15]*b.matrixdata[13];
	matrixdata[14]=a.matrixdata[12]*b.matrixdata[2]+a.matrixdata[13]*b.matrixdata[6]+a.matrixdata[14]*b.matrixdata[10]+a.matrixdata[15]*b.matrixdata[14];
	matrixdata[15]=a.matrixdata[12]*b.matrixdata[3]+a.matrixdata[13]*b.matrixdata[7]+a.matrixdata[14]*b.matrixdata[11]+a.matrixdata[15]*b.matrixdata[15];
}
 
template<class T> ostream &operator<<(ostream &output,const TransformMatrix<T> &matrix)
{
	for(size_t row=0;row < 4;row++)
	{
		for(size_t column=0; column < 4; column++)
		{
			output << /* setprecision(2) <<*/ matrix(row,column) << " ";
		}
		output << endl;
	}
	return output;
}

template<class T> template<class Vector> void TransformMatrix<T>::transformQuick(const Vector &v,Vector &vt) const
{
	vt[0]=v[0]*matrixdata[0]+v[1]*matrixdata[4]+v[2]*matrixdata[8]+matrixdata[12];
	vt[1]=v[0]*matrixdata[1]+v[1]*matrixdata[5]+v[2]*matrixdata[9]+matrixdata[13];
	vt[2]=v[0]*matrixdata[2]+v[1]*matrixdata[6]+v[2]*matrixdata[10]+matrixdata[14];
}

template<class T> template<class Vector> void TransformMatrix<T>::transform(const Vector &v,Vector &vt) const
{
	T normfactor=T(1.0)/(v[0]*matrixdata[3]+v[1]*matrixdata[7]+v[2]*matrixdata[11]+matrixdata[15]);
	vt[0]=(v[0]*matrixdata[0]+v[1]*matrixdata[4]+v[2]*matrixdata[8]+v[3]*matrixdata[12])*normfactor;
	vt[1]=(v[0]*matrixdata[1]+v[1]*matrixdata[5]+v[2]*matrixdata[9]+v[3]*matrixdata[13])*normfactor;
	vt[2]=(v[0]*matrixdata[2]+v[1]*matrixdata[6]+v[2]*matrixdata[10]+v[3]*matrixdata[14])*normfactor;
	
}

template<class T> template<class Vector> void TransformMatrix<T>::transformHomogen(const Vector &v,Vector &vt) const
{
	vt[0]=v[0]*matrixdata[0]+v[1]*matrixdata[4]+v[2]*matrixdata[8]+v[3]*matrixdata[12];
	vt[1]=v[0]*matrixdata[1]+v[1]*matrixdata[5]+v[2]*matrixdata[9]+v[3]*matrixdata[13];
	vt[2]=v[0]*matrixdata[2]+v[1]*matrixdata[6]+v[2]*matrixdata[10]+v[3]*matrixdata[14];
	vt[3]=v[0]*matrixdata[3]+v[1]*matrixdata[7]+v[2]*matrixdata[11]+v[3]*matrixdata[15];
}


template<class T> void TransformMatrix<T>::translate(T x, T y, T z)
{
	matrixdata[0]=matrixdata[0]+x*matrixdata[3];
	matrixdata[1]=matrixdata[1]+y*matrixdata[3];
	matrixdata[2]=matrixdata[2]+z*matrixdata[3];
	
	matrixdata[4]=matrixdata[4]+x*matrixdata[7];
	matrixdata[5]=matrixdata[5]+y*matrixdata[7];
	matrixdata[6]=matrixdata[6]+z*matrixdata[7];
	
	matrixdata[8]=matrixdata[8]+x*matrixdata[11];
	matrixdata[9]=matrixdata[9]+y*matrixdata[11];
	matrixdata[10]=matrixdata[10]+z*matrixdata[11];
	
	matrixdata[12]=matrixdata[12]+x*matrixdata[15];
	matrixdata[13]=matrixdata[13]+y*matrixdata[15];
	matrixdata[14]=matrixdata[14]+z*matrixdata[15];	
}

template<class T> void TransformMatrix<T>::rotateAroundZ(T phi)
{
	T c=cos(phi);
	T s=sin(phi);
	
	T a0=matrixdata[0];
	matrixdata[0]=a0*c-matrixdata[1]*s;
	matrixdata[1]=matrixdata[1]*c+a0*s;
	
	a0=matrixdata[4];
	matrixdata[4]=a0*c-matrixdata[5]*s;
	matrixdata[5]=matrixdata[5]*c+a0*s;
	
	a0=matrixdata[8];
	matrixdata[8]=a0*c-matrixdata[9]*s;
	matrixdata[9]=matrixdata[9]*c+a0*s;	
		
	a0=matrixdata[12];
	matrixdata[12]=a0*c-matrixdata[13]*s;
	matrixdata[13]=matrixdata[13]*c+a0*s;		
}

template<class T> void TransformMatrix<T>::rotateAroundX(T phi)
{
	T c=cos(phi);
	T s=sin(phi);
	
	T a1=matrixdata[1];
	matrixdata[1]=a1*c-matrixdata[2]*s;
	matrixdata[2]=matrixdata[2]*c+a1*s;
	
	a1=matrixdata[5];
	matrixdata[5]=a1*c-matrixdata[6]*s;
	matrixdata[6]=matrixdata[6]*c+a1*s;
	
	a1=matrixdata[9];
	matrixdata[9]=a1*c-matrixdata[10]*s;
	matrixdata[10]=matrixdata[10]*c+a1*s;	
		
	a1=matrixdata[13];
	matrixdata[13]=a1*c-matrixdata[14]*s;
	matrixdata[14]=matrixdata[14]*c+a1*s;		
}

template<class T> void TransformMatrix<T>::rotateAroundY(T phi)
{
	T c=cos(phi);
	T s=sin(phi);
	
	T a0=matrixdata[0];
	matrixdata[0]=a0*c+matrixdata[2]*s;
	matrixdata[2]=matrixdata[2]*c-a0*s;
	
	a0=matrixdata[4];
	matrixdata[4]=a0*c+matrixdata[6]*s;
	matrixdata[6]=matrixdata[6]*c-a0*s;
	
	a0=matrixdata[8];
	matrixdata[8]=a0*c+matrixdata[10]*s;
	matrixdata[10]=matrixdata[10]*c-a0*s;	
		
	a0=matrixdata[12];
	matrixdata[12]=a0*c+matrixdata[14]*s;
	matrixdata[14]=matrixdata[14]*c-a0*s;		
}

template<class T> void TransformMatrix<T>::scale(T sx, T sy, T sz)
{
	for(size_t k=0; k < 15; k+=4)
	{
		matrixdata[k] *= sx;
		matrixdata[k+1] *= sy;
		matrixdata[k+2] *=sz;	
	}
}

template<class T> void TransformMatrix<T>::projectionOrthoBottom()
{
	for(size_t k=1; k < 15; k+=4)
	{
		swap(matrixdata[k],matrixdata[k+1]);
	}
}

template<class T> void TransformMatrix<T>::projectionOrthoTop()
{
	for(size_t k=1; k < 15; k+=4)
	{
		T tmp=matrixdata[k];
		matrixdata[k]=matrixdata[k+1];
		matrixdata[k+1]=-tmp;
	}
}

template<class T> void TransformMatrix<T>::projectionOrthoLeft()
{
	for(size_t k=0; k < 15; k+=4)
	{
		T tmp=matrixdata[k];
		matrixdata[k]=matrixdata[k+2];
		matrixdata[k+2]=-tmp;
	}
}

template<class T> void TransformMatrix<T>::projectionOrthoRight()
{
	for(size_t k=0; k < 15; k+=4)
	{
		T tmp=matrixdata[k];
		matrixdata[k]=-matrixdata[k+2];
		matrixdata[k+2]=tmp;
	}
}

template<class T> void TransformMatrix<T>::projectionOnePointPersp(T d)
{
	T factor = T(1.0)/d;
	for(size_t k=3; k < 15; k+=4)
	{
		matrixdata[k]=matrixdata[k]-factor*matrixdata[k-1];
	}
}

/**
 * source code from intel's AP 962
 */
template<class T> void TransformMatrix<T>::invert()
{
  valarray<T> tmp(12); /* temp array for pairs */ 
  valarray<T> src(16); /* array of transpose source matrix */ 
  T det; /* determinant */ /* transpose matrix */ 
  for (size_t i = 0; i < 4; i++) 
    { 
      src[i] = matrixdata[i*4]; 
      src[i + 4] = matrixdata[i*4 + 1]; 
      src[i + 8] = matrixdata[i*4 + 2]; 
      src[i + 12] = matrixdata[i*4 + 3]; 
    } 
  /* calculate pairs for first 8 elements (cofactors) */ 
  tmp[0] = src[10] * src[15]; 
  tmp[1] = src[11] * src[14]; 
  tmp[2] = src[9] * src[15]; 
  tmp[3] = src[11] * src[13]; 
  tmp[4] = src[9] * src[14]; 
  tmp[5] = src[10] * src[13]; 
  tmp[6] = src[8] * src[15]; 
  tmp[7] = src[11] * src[12]; 
  tmp[8] = src[8] * src[14]; 
  tmp[9] = src[10] * src[12]; 
  tmp[10] = src[8] * src[13]; 
  tmp[11] = src[9] * src[12]; 
  /* calculate first 8 elements (cofactors) */ 
  matrixdata[0] = tmp[0]*src[5] + tmp[3]*src[6] + tmp[4]*src[7]; 
  matrixdata[0] -= tmp[1]*src[5] + tmp[2]*src[6] + tmp[5]*src[7]; 
  matrixdata[1] = tmp[1]*src[4] + tmp[6]*src[6] + tmp[9]*src[7]; 
  matrixdata[1] -= tmp[0]*src[4] + tmp[7]*src[6] + tmp[8]*src[7]; 
  matrixdata[2] = tmp[2]*src[4] + tmp[7]*src[5] + tmp[10]*src[7]; 
  matrixdata[2] -= tmp[3]*src[4] + tmp[6]*src[5] + tmp[11]*src[7]; 
  matrixdata[3] = tmp[5]*src[4] + tmp[8]*src[5] + tmp[11]*src[6]; 
  matrixdata[3] -= tmp[4]*src[4] + tmp[9]*src[5] + tmp[10]*src[6]; 
  matrixdata[4] = tmp[1]*src[1] + tmp[2]*src[2] + tmp[5]*src[3]; 
  matrixdata[4] -= tmp[0]*src[1] + tmp[3]*src[2] + tmp[4]*src[3]; 
  matrixdata[5] = tmp[0]*src[0] + tmp[7]*src[2] + tmp[8]*src[3]; 
  matrixdata[5] -= tmp[1]*src[0] + tmp[6]*src[2] + tmp[9]*src[3]; 
  matrixdata[6] = tmp[3]*src[0] + tmp[6]*src[1] + tmp[11]*src[3]; 
  matrixdata[6] -= tmp[2]*src[0] + tmp[7]*src[1] + tmp[10]*src[3]; 
  matrixdata[7] = tmp[4]*src[0] + tmp[9]*src[1] + tmp[10]*src[2]; 
  matrixdata[7] -= tmp[5]*src[0] + tmp[8]*src[1] + tmp[11]*src[2]; 
  /* calculate pairs for second 8 elements (cofactors) */ 
  tmp[0] = src[2]*src[7]; 
  tmp[1] = src[3]*src[6]; 
  tmp[2] = src[1]*src[7]; 
  tmp[3] = src[3]*src[5]; 
  tmp[4] = src[1]*src[6]; 
  tmp[5] = src[2]*src[5];
  tmp[6] = src[0]*src[7]; 
  tmp[7] = src[3]*src[4]; 
  tmp[8] = src[0]*src[6]; 
  tmp[9] = src[2]*src[4]; 
  tmp[10] = src[0]*src[5]; 
  tmp[11] = src[1]*src[4]; 
  /* calculate second 8 elements (cofactors) */ 
  matrixdata[8] = tmp[0]*src[13] + tmp[3]*src[14] + tmp[4]*src[15]; 
  matrixdata[8] -= tmp[1]*src[13] + tmp[2]*src[14] + tmp[5]*src[15]; 
  matrixdata[9] = tmp[1]*src[12] + tmp[6]*src[14] + tmp[9]*src[15]; 
  matrixdata[9] -= tmp[0]*src[12] + tmp[7]*src[14] + tmp[8]*src[15]; 
  matrixdata[10] = tmp[2]*src[12] + tmp[7]*src[13] + tmp[10]*src[15]; 
  matrixdata[10]-= tmp[3]*src[12] + tmp[6]*src[13] + tmp[11]*src[15]; 
  matrixdata[11] = tmp[5]*src[12] + tmp[8]*src[13] + tmp[11]*src[14]; 
  matrixdata[11]-= tmp[4]*src[12] + tmp[9]*src[13] + tmp[10]*src[14]; 
  matrixdata[12] = tmp[2]*src[10] + tmp[5]*src[11] + tmp[1]*src[9]; 
  matrixdata[12]-= tmp[4]*src[11] + tmp[0]*src[9] + tmp[3]*src[10]; 
  matrixdata[13] = tmp[8]*src[11] + tmp[0]*src[8] + tmp[7]*src[10]; 
  matrixdata[13]-= tmp[6]*src[10] + tmp[9]*src[11] + tmp[1]*src[8]; 
  matrixdata[14] = tmp[6]*src[9] + tmp[11]*src[11] + tmp[3]*src[8]; 
  matrixdata[14]-= tmp[10]*src[11] + tmp[2]*src[8] + tmp[7]*src[9]; 
  matrixdata[15] = tmp[10]*src[10] + tmp[4]*src[8] + tmp[9]*src[9]; 
  matrixdata[15]-= tmp[8]*src[9] + tmp[11]*src[10] + tmp[5]*src[8]; 
  /* calculate determinant */ 
  det=src[0]*matrixdata[0]+src[1]*matrixdata[1]+src[2]*matrixdata[2]+src[3]*matrixdata[3]; 
  /* calculate matrix inverse */ det = 1.0/det; 
  for(size_t j = 0; j < 16; j++) matrixdata[j] *= det; 
} 

#endif // TRANSFORMMATRIX_H
